# Generated by Django 3.2.8 on 2021-10-25 12:48

import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("djangostreetmap", "0005_osmislandsareas"),
    ]

    operations = [
        migrations.DeleteModel(
            name="PlanetOsmRoads",
        ),
        migrations.AlterField(
            model_name="osmislands",
            name="geom",
            field=django.contrib.gis.db.models.fields.MultiLineStringField(srid=3857),
        ),
    ]
